import numpy as np
import pytest
from packaging.utils import Version
import accelerated_numpy

old_numpy = Version(np.__version__) < Version('1.18')

@pytest.fixture(scope='session')
def initialize_accelerated_numpy():
    accelerated_numpy.initialize()

@pytest.fixture(scope='function')
def rng():
    if old_numpy:
        class OldRNG(np.random.RandomState):
            pass
        rng = OldRNG(1234)
        rng.random = rng.random_sample
        rng.integers = rng.randint
        return rng
    else:
        return np.random.default_rng(1234)



"""
Override all the NumPy UFunc loops with multithreaded variants. The original
loop will be called via a pool of threads.
"""
__version__ = 'v0.1.0'
__all__ = [
    'initialize', 'cpustring',
    'thread_enable', 'thread_disable', 'thread_isenabled', 'thread_getworkers', 'thread_setworkers',
    'ledger_enable', 'ledger_disable', 'ledger_isenabled', 'ledger_info',
    'recycler_enable', 'recycler_disable', 'recycler_isenabled', 'recycler_info',
    'timer_gettsc','timer_getutc']

from accelerated_numpy._accelerated_numpy import initialize, cpustring 
from accelerated_numpy._accelerated_numpy import thread_enable, thread_disable, thread_isenabled, thread_getworkers, thread_setworkers

import numpy as np

# TODO: move this to new location
def debug_timeit(func=np.equal, ctypes=[np.bool, np.int8, np.int16, np.int32, np.int64, np.float32, np.float64], scalar=False, unary = False, outdtype=None, recycle=True):
    timedelta = np.zeros(len(ctypes), np.int64)
    sizes=[1_000_000]
    for s in sizes:
        slot = 0
        loop_size = 100
        mtimedelta = np.zeros(loop_size, np.int64)
        for c in ctypes:
            if c is np.bool:
               a=np.arange(s, dtype=np.int8).astype(c)
            else:
               a=np.arange(s, dtype=c)
            if scalar is True:
                b=a[len(a)//2]
            else:
                b=a.copy()

            c = None
            if recycle:
                if outdtype is None:
                    c=np.zeros(s, dtype=c)
                else:
                    c=np.zeros(s, dtype=outdtype)

            for loop in range(loop_size):
                if unary is False:
                    starttime = fa.timer_gettsc()
                    func(a,b,out=c)
                    delta= fa.timer_gettsc() - starttime
                else:
                    starttime = fa.timer_gettsc()
                    func(a,out=c)
                    delta= fa.timer_gettsc() - starttime
                mtimedelta[loop] = delta
            timedelta[slot] = np.median(mtimedelta)
            slot = slot + 1
    return timedelta


def debug_benchmark(
    func=np.equal,
    ctypes=[np.bool, np.int8, np.int16, np.int32, np.int64, np.float32, np.float64],
    scalar=False,
    unary = False,
    outdtype=None,
    recycle=True,
    atop=True,
    thread=True):

    # fa.atop_disable()
    fa.thread_disable()
    # get original time
    t0=debug_timeit(func=func, ctypes=ctypes, scalar=scalar, unary=unary, outdtype=outdtype, recycle=recycle)
    # if atop:
    #    fa.atop_enable()
    if thread:
        fa.thread_enable()
    t1=debug_timeit(func=func, ctypes=ctypes, scalar=scalar, unary=unary, outdtype=outdtype, recycle=recycle)
    return t0/t1

def benchmark(
    ctypes=[np.bool, np.int8, np.int16, np.int32, np.int64, np.float32, np.float64],
    recycle=True,
    atop=True,
    thread=True):

    print("equal twoinp", debug_benchmark(np.equal, ctypes=ctypes, scalar=False, unary=False, recycle=recycle, atop=atop, thread=thread,  outdtype='?'))
    print("equal scalar", debug_benchmark(np.equal, ctypes=ctypes, scalar=True, unary=False, recycle=recycle, atop=atop, thread=thread, outdtype='?'))
    print("add   twoinp", debug_benchmark(np.add, ctypes=ctypes, scalar=False, unary=False,recycle=recycle, atop=atop, thread=thread))
    print("add   scalar", debug_benchmark(np.add, ctypes=ctypes, scalar=True, unary=False,recycle=recycle, atop=atop, thread=thread))
    print("abs         ", debug_benchmark(np.abs, ctypes=ctypes, scalar=False, unary=True,recycle=recycle, atop=atop, thread=thread))
    print("isnan       ", debug_benchmark(np.abs, ctypes=ctypes, scalar=False, unary=True,recycle=recycle, atop=atop, thread=thread))




import numpy as np
import accelerated_numpy as fn
fn.initialize()

def test_enable():
    # enable/disable return the previous value
    old = fn.thread_isenabled()
    fn.thread_enable()
    assert fn.thread_isenabled() == True
    fn.thread_disable()
    assert fn.thread_isenabled() == False

    # restore prior state
    if old:
        fn.thread_enable()
    else:
        fn.thread_disable()
    assert fn.thread_isenabled() == old

def test_result(rng):
    print('numpy version', np.__version__)
    print(fn.cpustring())
  
    m = rng.integers(100, size=(10, 10), dtype=np.int32)
    o = np.empty_like(m)
    for i in range(m.shape[0]):
        for j in range(m.shape[1]):
            o[i, j] = m[i, j] + m[i, j]
    assert np.all(np.add(m, m) == o)

def test_numpy():
    np.test()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import os
import platform
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import relpath
from os.path import splitext

from setuptools import Extension
from setuptools import find_packages
from setuptools import setup
import numpy as np

# Enable code coverage for C code: we can't use CFLAGS=-coverage in tox.ini, since that may mess with compiling
# dependencies (e.g. numpy). Therefore we set SETUPPY_CFLAGS=-coverage in tox.ini and copy it to CFLAGS here (after
# deps have been safely installed).
if 'TOX_ENV_NAME' in os.environ and os.environ.get('SETUP_PY_EXT_COVERAGE') == 'yes' and platform.system() == 'Linux':
    CFLAGS = os.environ['CFLAGS'] = '-fprofile-arcs -ftest-coverage -DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION'
    LFLAGS = os.environ['LFLAGS'] = '-lgcov'
else:
    CFLAGS = '-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION -DPy_LIMITED_API'
    LFLAGS = ''

if platform.system() == 'Windows':
    CFLAGS += ' /Ox /Ob2 /Oi /Ot /d2FH4-'
else:
    CFLAGS += ' -mavx2 -fpermissive -Wno-unused-variable -Wno-unused-function -std=c++11 -pthread -falign-functions=32'

if platform.system() == 'Linux':
    LFLAGS += ' -lm'

def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


with open("README.md") as readme:
    long_description = readme.read()

import _add_newdocs
_add_newdocs.main()

setup(
    name='accelerated-numpy',
    version='v0.1.0',
    license='MIT',
    description='Faster loops for NumPy using multithreading and other tricks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Matti Picus',
    author_email='mattigit@picus.org.il',
    url='https://github.com/Quansight/numpy-threading-extensions',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    project_urls={
        'Changelog': 'https://github.com/Quansight/numpy-threading-extensions/blob/master/CHANGELOG.rst',
        'Issue Tracker': 'https://github.com/Quansight/numpy-threading-extensions/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.6',
    install_requires=[
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
        'numpy>=1.19.0',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    ext_modules=[
        Extension(
            'accelerated_numpy._accelerated_numpy',
            sources=['src/accelerated_numpy/_accelerated_numpy.cpp',
                     'src/accelerated_numpy/module_init.cpp',
                     'src/accelerated_numpy/ledger.cpp', 
                     'src/atop/atop.cpp',
                     'src/atop/threads.cpp',
                     'src/atop/ops_binary.cpp',
                     'src/atop/ops_compare.cpp',
                     'src/atop/ops_unary.cpp',
                     'src/atop/ops_trig.cpp',
                     'src/atop/ops_log.cpp',
                    ],
            extra_compile_args=CFLAGS.split(),
            extra_link_args=LFLAGS.split(),
            include_dirs=['src/accelerated_numpy', 'src/atop', np.get_include()],
            py_limited_api=True,
        )
    ],
)
